package CTL.Comm;

import CTL.*;
import CTL.Types.*;

import com.jcraft.jsch.*;
import java.io.*;

/** Wrapper for the jsch SSHv2 API */
public class SSHv2 implements Runnable
{
	/** Command to invoke */
	private String start;
	/** Location of the remote machine */
	private Location loc;
	/** Communicator which uses this transport */
	private Communicator comm;

	/** 
	 * Check if the user has a RSA private key for SSH connections
	 * @return True if he does, false otherwise
	 */
	public static boolean hasPrivateKey ()
	{
		return new File(System.getProperty("user.home")+"/.ssh/id_rsa").exists();
	}
	
	/** Initialization
	 * @param loc Location
	 */
	public SSHv2 (Location loc, Communicator comm, Env env)
	{
		this.comm = comm;
		this.loc = loc;
		this.start = Globals.GetTrace ? "strace -f -o /tmp/trace.log " : "";
		//System.out.println(comm.pid());
		this.start += "env "+loc.path()+"/"+loc.cmd()+" -l tcp -h "+
			comm.pid().host()+" -S "+env.grp.members()+" -f /tmp/service.log"+
			" -v -L "+comm.pid().port()+" -p "+(env.grp.myID()+1);
		env.log.msg(Logger.DBG2, "Starting service with "+this.start);
		if (env.debugger != null)
			start += " -D "+env.debugger.host()+":"+env.debugger.port();
	}

	/** Runs the command specified by the Location on the remote machine */
	public void run ()
	{
		try
		{
			if ((loc==null)||(loc.path()==null)||(loc.cmd()==null))
			{
				Env.log.msg(Logger.ERR, "Location not initialized!");
				throw new RuntimeException("No location to connect to.");
			}

			if (Env.debugger != null)
			{
				BuffyOut out = new BuffyOut();
				PeerID pid = new PeerID(loc.host(), loc.port());
				Header head = new Header(pid.size(), Remote.SSH, comm.pid(), -1);
				out.o.serialWrite(head);
				out.o.serialWrite(pid);
				out.o.flush();
				Remote.writeDBG(comm, out.getBytes());
			}
			
			JSch jsch = new JSch();
			if (hasPrivateKey())
				jsch.addIdentity(System.getProperty("user.home")+"/.ssh/id_rsa");
			String hosts = System.getProperty("user.home")+"/.ssh/known_hosts";
			if (new File(hosts).exists())
				jsch.setKnownHosts(hosts);
			else
				Env.log.msg(Logger.ERR, hosts+": No such file or directory.");
		
			Env.log.msg(Logger.INFO, loc.toString());
			if (loc.user==null)
				loc.user=System.getProperty("user.name");
			Session sess = jsch.getSession(loc.user, loc.host(), 
				loc.port());
			loc.setPassword(sess);
			sess.connect();

			Channel chan = sess.openChannel("exec");
			if (chan==null)
			{
				Env.log.msg(Logger.ERR, "Session could not be opened.");
				throw new RuntimeException("SSH connection refused.");
			}

			((ChannelExec)chan).setCommand(start);
			chan.connect();
			if (chan.isEOF())
				throw new RuntimeException("SSH connection reset.");
			chan.setOutputStream(System.out);
			//chan.setInputStream(System.in);
			((ChannelExec)chan).setErrStream(System.err);
			Env.log.msg(Logger.INFO, "Service spawned via SSH.");

			while (!chan.isEOF())
				Thread.sleep(500);
			sess.disconnect();	
		}
		catch (Exception e) 
		{
			Env.log.msg(Logger.ERR, RUtil.stackTrace(e));
		}
		finally
		{
			if (Env.clean)
				return;
			System.out.println("Unexpected termination of remote service.");
			comm.notifyErr();
		}
	}
}
