package CTL.Types;

import CTL.*;
import CTL.Serialize.*;
import CTL.Streams.*;
import ReflWrap.*;

import java.io.*;
import java.lang.reflect.*;

/** CTL Tupel (a fixed-sized array of multiple types) */
// TODO: variable number of type parameters
public class Tupel<A,B> implements Writable, TemplHack
{
	/** Type paramenters */
	protected TypeTree[] type;
	/** Stored data */
	protected Object[] data;

	/** Array helper function
	 * @param array Array of classes
	 * @param idx Index number
	 * @param moo New value 
	 * @return Array of classes with the specified value replaced
	 */
	protected static Class[] insert (Class[] array, int idx, Class moo)
	{
		array[idx] = moo;
		return array;
	}

	/** Set the type parameters
	 * @param types Array of classes
	 */
	public void setTypes (TypeTree[] types)
	{
		//System.out.println("Setting types to: "+Arrays.toString(types));
		type = types;
	}

	/** Retrieve a string representation of this object
	 * @return String
	 */
	public String toString ()
	{
		StringBuffer buf = new StringBuffer();
		buf.append("Tupel (ID #"+hashCode()+"):\n");
		for (int i=0;i<type.length;i++)
			buf.append("\t"+type[i]+": "+data[i]+
				((i!=type.length-1) ? "\n" : ""));
		return buf.toString();
	}

	/** Dummy constructor to make subclasses happy */
	protected Tupel ()
	{
	}

	/** Generate a new Tupel
	 * @param type Type parameters
	 */
	public Tupel (Class[] type) throws CTLException, ClassNotFoundException
	{
		if (type == null)
			throw new CTLException("Invalid Tupel.");

		this.type = new TypeTree[type.length];
		this.data = new Object[type.length];
		for (int i=0;i<type.length;i++)
			this.type[i] = new TypeTree(type[i]);
	}

	/** Constructor from TypeTree */
	public Tupel (TypeTree[] tree) throws CTLException, ClassNotFoundException
	{
		if (tree == null)
			throw new CTLException("Invalid Tupel.");

		//System.out.println(java.util.Arrays.toString(tree));
		this.type = tree.clone();
		this.data = new Object[tree.length];
	}

	/** Retrieve the number of elements this Tupel can store
	 * @return Number of elements
	 */
	public int length () 
	{
		return data.length;
	}

	/** Retrieve the type of a specific element 
	 * @param i Index number
	 * @return Type of the element
	 */
	public Class type (int i) throws CTLException
	{
		if ((i<0)||(i>=type.length))
			throw new CTLException(i+" is out of bounds.");
		return type[i].getType();
	}

	/** Retrieve the value of a specific element
	 * @param i Index number
	 * @return Value of the element
	 */
	public Object item (int i) throws CTLException
	{
		if ((i<0)||(i>=data.length))
			throw new CTLException(i+ "is out of bounds.");
		return data[i];
	}

	/** Set the value of a specific element
	 * @param i Index number
	 * @param data New value
	 */
	public void setItem (int i, Object data) throws CTLException
	{
		if (data!=null)
		{
			if ((i<0)||(i>=this.type.length))
				throw new CTLException(i+ " is out of bounds.");

			ClassInfo c1 = new ClassInfo(type[i].getType());
			ClassInfo c2 = new ClassInfo(data.getClass());
			
			if (!c1.equals(c2))
				throw new CTLException("Type mismatch: "+c1.name()+
					" != "+c2.name());
		}		
		
		//System.out.println(data);
		this.data[i] = data;
	}
	
	/** Serial read function
	 * @param in Input stream
	 */
	public void read (SerialIn in) throws IOException, ClassNotFoundException, 
		InstantiationException, IllegalAccessException, InvocationTargetException
	{
		for (int i=0;i<type.length;i++)
			data[i] = IStream.readType(in, type[i].getType());
	}
	
	/** Serial write function
	 * @param out Output stream
	 */
	public void write (SerialOut out) throws IOException, IllegalAccessException,
		InvocationTargetException
	{
		for (int i=0;i<data.length;i++)
			IStream.writeType(out, data[i]);
	}

	/** Check if two objects are equal
	 * @param t Object to compare to
	 * @return True if equal, false otherwise
	 */
	public boolean equals (Object t)
	{
		if (!(t instanceof Tupel) || t==null)
			return false;
		
		try
		{
			if (length()!=((Tupel)t).length())
				return false;
			for (int i=0;i<length();i++)
				if ((item(i) == null && ((Tupel)t).item(i) == null) || 
					(!item(i).equals(((Tupel)t).item(i))))
					return false;
			return true;
		}
		catch (CTLException e)
		{
			Env.log.msg(Logger.ERR, "Tupel.equals(): "+e);
			return false;
		}
	}
}
