package CodeGen;

import CTL.*;
import ReflWrap.*;

import java.io.*;
import java.lang.annotation.*;
import java.lang.reflect.*;
import java.util.*;

/** Abstract code generator interface */
public abstract class Gen
{
	public ClassInfo cinfo;
	protected Konstruktor[] constr;
	protected Methode[] methods;
	protected String name, pkg, prfx, supername;
	protected String modifiers = "";
	protected boolean isForwardDecl = false, isCTLI = false;

	/** Constructor for a class  from the javaSys package.
	 * @param klass Class name
	 */
	public Gen (String klass) throws ClassNotFoundException
	{
		this("javaSys", klass);
	}

	/** Constructor for a class  from a custom package.
	 * @param pkg Package name
	 * @param klass Class name
	 */
	public Gen (String pkg, String klass) throws ClassNotFoundException
	{
		this.pkg = pkg;
		cinfo = new ClassInfo(Class.forName(klass));
		constr = Konstruktor.fetchAll(cinfo);
		methods = Methode.fetchAll(cinfo);

		if (cinfo._package().equals("_default"))
			this.pkg = "_default";

		if (this.pkg.equals("_default"))
			prfx = "";
		else
			prfx = this.pkg+"::";
	}

	/** Write a string to an output stream
	 * @param out OutputStreamWriter for the output stream
	 * @param str String to write
	 */
	protected static void writeTo (OutputStreamWriter out, String str) 
		throws IOException
	{
		out.write(str, 0, str.length());
	}

	/** Convert a class name to a Java source code filename
	 * @param klas Class name
	 * @return Filename
	 */
	protected static String classNameToDotJavaFile (String klas)
	{
		StringBuffer buf = new StringBuffer();
		StringTokenizer str = new StringTokenizer(klas, ".");

		while (str.hasMoreTokens())
			buf.append(str.nextToken()+((str.hasMoreTokens()) ? "/" : ""));
		buf.append(".java");
			
		return "src/"+buf.toString();
	}

	/** Retrieve the import statements from a Java source code file
	 * @param klas Class name
	 * @param buf Buffer to write to
	 */
	protected static void getImports (StringBuffer buf, String klas)
	{
		try
		{
			FileReader fin = new FileReader(classNameToDotJavaFile(klas));
			LineNumberReader lin = new LineNumberReader(fin);

			String line = lin.readLine();
			while (line!=null)
			{
				if ((line.length()>5) && (line.substring(0, 6).equals("import")))
					buf.append(line+"\n");
				line = lin.readLine();
			}

			fin.close();
		} catch (IOException e) {}
	}

	/** Retrieve all imported classes/packages as String
	 * @param klas Name of the source class
	 * @return String
	 */
	protected static String getImports (String klas)
	{
		StringBuffer buf = new StringBuffer();
		getImports(buf, klas);
		return buf.toString();
	}

	/** Use an external pretty printer
	 * @param file The File to modify
	 */
	protected static void prettyPrint (File file)
	{
		try
		{
			Runtime rt = Runtime.getRuntime();
			java.lang.Process proc = rt.exec("astyle "+file.toString());
			if (proc.waitFor()==0)
			{
				File orig = new File(file.toString()+".orig");
				orig.delete();
			}
		} catch (Exception e) {}
	}			

	protected String classNamer (ClassInfo cinfo, String suffix)
	{
		String name = cinfo.name();
		if (ByteCode.matches(name, ".*CTLI$"))
		{
			isCTLI = true;
			return name.substring(0, name.length()-4);
		}
		if (ByteCode.matches(name, ".*ForwardDecl"))
		{
			isForwardDecl = true;
			return name.substring(0, name.length()-"ForwardDecl".length())+suffix;
		}
		return name+suffix;
	}

	protected boolean isCC () 
	{
		String foo = classNamer(cinfo, "RI");
		String bar = classNamer(cinfo, "CI");
		return foo.equals(bar);
	}

	protected void generate () throws IOException, NoSuchMethodException
	{
		FileOutputStream fos = null;

		try
		{
			// Check if the file was modified recently
			File impl = new File(cinfo.filename());
			String fname = pkg+"/"+name+".java";
			if (new File("src").exists())
				fname = "src/"+fname;
			if (!new File("src/"+pkg).exists())
			{
				if (new File("src").exists())
					new File("src/"+pkg).mkdir();
				else
					new File(pkg).mkdir();
			}
			File ri = new File(fname);

			if (impl.lastModified()<ri.lastModified())
			{
				File gen = new File("src/CodeGen/GenRI.java");
				if (gen.lastModified()<ri.lastModified())
					return;
			}

			ri.delete();
			fos = new FileOutputStream(ri);
			OutputStreamWriter ostr = new OutputStreamWriter(fos);
			gen(ostr);
			ostr.flush();
			prettyPrint(ri);
		}
		finally
		{
			if (fos != null)
				fos.close();
		}
	}

	protected void gen (OutputStreamWriter ostr) throws IOException, NoSuchMethodException
	{
		writeTo(ostr, genHead());
		writeTo(ostr, genConstructors());
		writeTo(ostr, genWrappers());
		writeTo(ostr, genAccept());
		writeTo(ostr, "}\n");
	}

	/** Generate the header of the class
     * @return Header
     */
    protected String genHead () throws IOException
    {
        StringBuffer ret = new StringBuffer();

        ret.append("package "+pkg+";\n\n");
        ret.append("import CTL.*;\n");
        ret.append("import CTL.CCompat.*;\n");
        ret.append("import CTL.Comm.*;\n");
        ret.append("import CTL.Streams.*;\n");
        ret.append("import CTL.Types.*;\n");
        ret.append("import ReflWrap.*;\n\n");
        ret.append("import java.io.*;\n");
        ret.append("import java.net.*;\n");
        ret.append("import java.lang.reflect.*;\n");
		ret.append("import java.util.*;\n");
        if (cinfo._package().equals(""))
        {
            System.err.println("Implementation belongs to no package.");
            System.exit(10);
        }
        else
            ret.append("import "+cinfo._package()+".*;\n");
        getImports(ret, cinfo.fqcn());

        ret.append("\n/* This class was autogenerated by the CTL4j code"+
            " generator from\n *\t\t"+getFQCN()+".\n * Written by Boris"+
            " Buegling, licensed under the GNU General Public license. */\n");
        
		ret.append("public "+modifiers+" class "+name+" extends "+supername+"\n{\n");
		ret.append("public String getBase ()\n{\n");
		ret.append("return \""+getFQCN()+"\";\n}\n");
        return ret.toString();
    }		

	/** Generate a *Value() call for a certain type
     * @param retVal String representation of a type
     * @return Array of which the first String is the *Value() function
     * name
     */
    protected static String[] genValue (String retVal)
    {
        String[] ret = new String[2];
        Class foo = Refl.getClass(Refl.convert(retVal));

        if (retVal.equals("Array"))
            retVal = "Object";

        ret[1] = "";

        if ((retVal!=null)&&(foo==null))
        {
            ret[1] = "("+retVal+")";
            retVal = "Object";
        }

        ret[0] = retVal+"Value";

        return ret;
    }

	protected String genAccept () throws IOException
	{
		String foo = supername.equals("RI") ? name : supername;
		return "public static void accept (OIStream in, Header ohead,"+
			" long objID, FID fid, int id, Env env) throws IOException,"+
			" ClassNotFoundException, InstantiationException,"+
			" IllegalAccessException, NoSuchMethodException,"+
			" InvocationTargetException, CTLException\n{\n"+
			foo+".accept(in, ohead, objID, fid, id, env);\n}\n";
	}

	protected String genConstructors () throws IOException
	{
		StringBuffer ret = new StringBuffer();
		
		if (constr.length == 0)
		{
			ret.append("public "+name+ " ()\n{\n");
			if (!supername.equals("RI"))
				ret.append("super((Object)null);\n");
			ret.append("}\n");
		}

		for (Konstruktor con : constr)
		{
			ret.append("public "+name+" ("+con.params()+"\n{\n");
			if (!supername.equals("RI"))
				ret.append("super((Object)null);\n");
			ret.append("}\n");
		}

		return ret.toString();
	}

	protected String genWrappers () throws IOException, NoSuchMethodException
	{
		StringBuffer ret = new StringBuffer();

		for (Methode meth : methods)
		{
			if ((meth == null) || (!meth.here()) || (meth.dont()))
				continue;
			ret.append(meth.head(false)+"\n{\n");
			if (!meth.isVoid())
				ret.append("return "+Refl.defaultVal(meth.returns())+";\n");
			ret.append("}\n");
		}

		return ret.toString();
	}

	/** Generate code for linkage error message
     **/
    protected String linkErr ()
    {
        return "System.out.println(\"Critical error, no matching linkage for \"+"+
            "Location.linkage2str(proc().loc().linkage())+\".\");\n";
    }

    /** Generate code for allocating a new RI object
     * @param clname Class name
     * @param arglist Argument list
     * @param have_proc Put proc into the argument list or not
     * @return Code
     **/
    protected static String newObj (String clname, String arglist, boolean have_proc)
    {
        StringBuffer con = new StringBuffer();
        con.append("self = new "+clname+"(");
        if (have_proc)
            con.append("proc");
        if (!arglist.equals(""))
        {
            if (have_proc)
                con.append(", ");
            con.append(arglist);
        }
        con.append(");\n");
		con.append("return;\n");
        return con.toString();
    }

	/** See above */
    protected static String newObj (String clname, String arglist)
    {
        return newObj(clname, arglist, true);
    }

	/** Generate a stub for putRefCount()
     * @return Code
     */
    protected static String __putRefCount ()
    {
        StringBuffer buf = new StringBuffer();
        buf.append("case 0:\n");
        buf.append("int arg0 = in.readInt();\n");
        buf.append("env.log.msg(Logger.ERR, \"STUB: RI.putRefCount()"+
            " (Value: \"+arg0+\")\");\n");
        buf.append("break;\n");
        return buf.toString();
    }

	/** Handles one function parameter and takes care of possible array
	 * types.
	 * @param buf StringBuffer to write to
	 * @param p Parameter type
	 * @param i Counter
	 * @param j Counter
	 * @param is_constr Whether or not the caller is a constructor
	 * @param anot Annotations belonging to the caller
	 */
	protected static void genArgList (StringBuffer buf, String p, int i, int j, 
		boolean is_constr, Annotation[] anot)
	{
		String reader = RUtil.getReader(p).getName();
		boolean is_arr = reader.equals("readArray");
		boolean foobar = p.endsWith("[]");

		// TODO: Ugly hack
		if (foobar)
		{
			String foo = p.substring(0, p.length()-2);
			String bar = convertArray(foo);
			if (bar.equals("("+foo+"[])"))
			{
				foobar = false;
				is_arr = true;
			}
		}

		if (!Refl.hasAnot("@CTL.Any()", anot) && reader.equals("readObject"))
			reader="serialRead";
		
		if (is_constr)
			buf.append("foo"+i+"["+j+"] = ");
		else
			buf.append(((is_arr) ? "Object[]" : p)+" arg"+((is_arr) ? "T" : "")+
			j+i+" = ");
		
		if (!is_arr)
			buf.append(((!Refl.convert(p).equals(RUtil.reader2str(reader))) ? "("+p+")" : ""));

		if (foobar)
		{
			String foo = p.substring(0, p.length()-2);
			buf.append(convertArray(foo)+"(");
		}
		
		buf.append("in."+reader+"(");
		
		/* **** Seems to be old code **** */
		//buf.append((is_arr) ? Refl.arrayType2(p) : "");
		// TODO: Temporary hack, does not work for parametrized types
		//buf.append((is_arr) ? ", null" : "");
			
		if (reader.equals("serialRead"))
		{
			String type = Refl.firstToken(p, "<");
			if (!type.equals(p))
				buf.append("new TypeTree(\""+p+"\")");
			else
				buf.append(p+".class");
		}
		
		if (foobar)
			buf.append(")");
		buf.append(");\n");
		
		// TODO: This will not work for constructors
		if (is_constr) return;
		if (is_arr)
		{
			buf.append(p+" arg"+j+i+" = new "+p.substring(0, p.length()-1)+
				"argT"+j+i+".length];\n");
			buf.append("for (int i=0;i<argT"+j+i+".length;i++)\n");
			buf.append("arg"+j+i+"[i] = "+RUtil.converter(p.substring(0,
				p.length()-2), "argT"+j+i+"[i]")+";\n");
		}
	}
	
	/** Generate code for a new Method call
     * @param clname Class name
     * @param funcname Function name
     * @param len Number of arguments the function takes
	 * @param returnme Whether or not to return a value
     * @return Code
     **/
    protected static String newMethod (String clname, String funcname, int len, 
		boolean returnme)
    {
        StringBuffer ret = new StringBuffer();
        ret.append((returnme ? "return " : "")+clname+"."+funcname+"(");
        for (int j=0;j<len;j++)
            ret.append("arg"+j+((j==len-1) ? "" : ", "));
        ret.append(");\n");
        return ret.toString();
    }

	protected static String newMethod (String clname, String funcname, int len)
	{
		return newMethod(clname, funcname, len, true);
	}

	/** Retrieve the full-qualified name of the class we are generating code for
	 * @return Name
	 */
	protected String getFQCN ()
	{
		String name = cinfo.fqcn();
		if (isForwardDecl)
			name = name.substring(0, name.length()-"ForwardDecl".length());
		return name;
	}

	/** Wrapper around convertArray(String)
	 * @param minfo The method whose return value should be converted
	 * @return Array conversion expression
	 */
	protected static String convertArray (MethodInfo minfo)
	{
		return convertArray(minfo.returns().arrayType().getSimpleName());
	}

	/** Generate an array conversion expression for the given class
	 * @param klass Class name
	 * @return Array conversion expression
	 */
	protected static String convertArray (String klass)
	{
		String method = klass+"Array";
		Class args[] = new Class[1];
		args[0] = Object.class;
		try
		{
			Method foo = Refl.class.getDeclaredMethod(method, args);
			return "Refl."+method;
		}
		catch (NoSuchMethodException e) {}
		//return "("+klass+"[])Refl.ObjectArray";
		return "("+klass+"[])";
	}

	/** Generated code for the conversion of an array
	 * @param type Type of the array
	 * @return Generated code
	 */
	protected static String arrayConversion (String type)
	{
		StringBuffer buf = new StringBuffer();
		buf.append("Object[] tmp = (Object[])data;\n");
		buf.append(type+"[] tmp2 = new "+type+"[tmp.length];\n");
		buf.append("for (int i=0;i<tmp.length;i++)\n");
		buf.append("tmp2[i] = ("+type+")tmp[i];\n");
		return buf.toString();
	}		
}
