package ReflWrap;

import java.lang.annotation.*;
import java.lang.reflect.*;
import java.util.*;

/** Wrapper around java.lang.reflect.Method
 * Retrieve information about methods at runtime
 */
public class MethodInfo
{
	/** The Method object */
	private Method m;
	/** Name of the method */
	private String name;
	/** Types of this method's parameters */
	private String params[];

	/** Get the declaring class of this method
	 * @return Declaring class
	 */
	public Class _class ()
	{
		return m.getDeclaringClass();
	}

	/** Get the Method object itself
	 * @return Method object
	 */
	public Method m ()
	{
		return m;
	}

	/** Get the name of the method
	 * @return Name
	 */
	public String name ()
	{
		return name;
	}

	/** Get the return value type of the method
	 * @return Return value
	 */
	public ClassInfo returns ()
	{
		return new ClassInfo(m.getReturnType());
	}

	/** Determines the position of this method in the Method array of BCEL
	 * @return Position
	 */
	private int determineID ()
	{
		LinkedList<String> bcel = ByteCode.getMethods(_class().getName());
		String method = toSimpleString();
		//System.out.println(bcel);
		//System.out.println(method);
		/*for (int i=0;i<bcel.size();i++)
			System.out.println(i+" _"+bcel.get(i)+"_");
		System.out.println("\nLooking for: _"+method+"_");*/
		return bcel.indexOf(method);
	}

	/** Return value with template parameters
	 * @return String
	 */
	public String returnsT ()
	{
		//System.out.println(determineID());
		TypeTree ret = ByteCode.getRetVal(_class().getName(), 
			determineID());
		return (ret == null ? returns().name() : ret.toString());
	}
			
	/** Get parameter types of this method
	 * @return Array of parameter types
	 */
	public String[] params ()
	{
		return params;
	}

	/** Get parameter types of this method
	 * @return Array of parameter types
	 */
	public Class[] getParams ()
	{
		return m.getParameterTypes();
	}
	
	/** Constructor from a Method object
	 * @param method Existing Method object
	 */
	public MethodInfo (Method method)
	{
		m=method;
		name=m.getName();

		Type p[] = m.getGenericParameterTypes();
		params = new String[p.length];
		LinkedList<TypeTree> list = ByteCode.getParameters(
			_class().getName(), determineID()); 
		
		for (int i=0;i<p.length;i++)
		{
			if (list.size() > 0)
				params[i] = list.get(i).toString();
			else
				params[i] = Refl.printType(p[i]);
		}
	}
	
	/** Get a string representation of this object
	 * @return String
	 */
	public String toString ()
	{
		StringBuffer ret = new StringBuffer();
		
		ret.append(modifiersToString());
		if (returns().isArray())
			ret.append(returns().arrayType().getName()+"[]");
		else
			ret.append(returnsT()); //.name());
		ret.append(" "+name+" (");
		ret.append(paramsToString());
		
		return ret.toString();
	}

	/** Convert all parameters to a single string
	 * @return String with all parameters
	 */
	public String paramsToString ()
	{
		StringBuffer ret = new StringBuffer();
		int i;
		
		for (i=0;i<params.length;i++)
			ret.append(params[i]+" arg"+i+ 
					((i==params.length-1) ? ")" : ", "));
			
		if (i==0)
			ret.append(")");

		return ret.toString();
	}

	/** Checks whether this method is static
	 * @return True if static, false if not
	 */
	public boolean isStatic ()
	{
		int mod = m.getModifiers();

		if ((mod & Modifier.STATIC) != 0)
			return true;

		return false;
	}	
	
	/** Convert all modifiers of this method to a string
	 * @return String containing all modifiers
	 */
	public String modifiersToString ()
	{
		StringBuffer sb = new StringBuffer();
		int mod = m.getModifiers();

		if ((mod & Modifier.PUBLIC) != 0)
			sb.append("public ");
		if ((mod & Modifier.PRIVATE) != 0)
			sb.append("private ");
		if ((mod & Modifier.PROTECTED) != 0)
			sb.append("protected ");
		if ((mod & Modifier.ABSTRACT) != 0)
			sb.append("abstract ");
		if ((mod & Modifier.STATIC) != 0)
			sb.append("static ");
		if ((mod & Modifier.FINAL) != 0)
			sb.append("final ");
		if ((mod & Modifier.TRANSIENT) != 0)
			sb.append("transient ");
		if ((mod & Modifier.VOLATILE) != 0)
			sb.append("volatile ");
		if ((mod & Modifier.NATIVE) != 0)
			sb.append("native ");
		if ((mod & Modifier.SYNCHRONIZED) != 0)
			sb.append("synchronized ");
		if ((mod & Modifier.INTERFACE) != 0)
			sb.append("interface ");

		return sb.toString();
	}

	/** Get all annotations associated with a parameter
	 * @param i Index number of the parameter
	 * @return Array of Annotation objects
	 */
	public Annotation[] getParamAnot (int i)
	{
		Annotation[][] aa = m.getParameterAnnotations();
		return aa[i];
	}

	/** Returns this element's annotation for the specified type if such an
	 * annotation is present, else null
	 * @param klass Specified type
	 * @return Element's annotation
	 */
	public <T extends Annotation> T anot (Class<T> klass)
	{
		return m.getAnnotation(klass);
	}

	/** Returns a string representation of this object which can be 
	 * compared to strings retrieved by BCEL
	 * @return String
	 */
	public String toSimpleString ()
	{
		String s = m.toString();
		s = ByteCode.sed(s, ",", ", ");
		s = ByteCode.sed(s, " arg[0-9]*", "");		
		s = ByteCode.sed(s, "<.*?>", "");
		s = ByteCode.sed(s, "\\(", " ( ");
		s = ByteCode.sed(s, "\\)", ")");
		s = ByteCode.unclassify(s);
		s = ByteCode.sed(s, "\\( ", "(");
		s = ByteCode.sed(s, " \\) $", ")");
		s = ByteCode.sed(s, "\\) $", ")");
		s = ByteCode.sed(s, " throws.*", "");
		return s;
	}
}
